/************************************************************************************
Blackship RT-Component
Copyright (c) 2010, Segway-Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/


#include <stdio.h>
#ifndef WIN32
#include <math.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <sys/poll.h>
#else
#include "winterm.h"
#endif

#include "Blackship.h"
#include "bs_if.h"

#define BLACKSHIP_BAUDRATE    B19200


#define BLACKSHIP_STX               ((unsigned char)0x02)
#define BLACKSHIP_ETX               ((unsigned char)0x03)

#define BLACKSHIP_CPUID_MOTOR_RIGHT ((unsigned char)0x01)
#define BLACKSHIP_CPUID_MOTOR_LEFT  ((unsigned char)0x02)
#define BLACKSHIP_CPUID_MASTER      ((unsigned char)0x99)

#define BLACKSHIP_MODE_MOTOR_TIMEOUT    ((unsigned char)0x08)
#define BLACKSHIP_MODE_MOTOR_STATUS     ((unsigned char)0x09)
#define BLACKSHIP_MODE_MOTOR_REQUEST    ((unsigned char)0xF9)

#define BLACKSHIP_MODE_MOTOR_SETSPEED   ((unsigned char)0x00)
#define BLACKSHIP_MODE_MOTOR_SETTIMEOUT ((unsigned char)0x07)

#define BLACKSHIP_MAX_MSG_LEN       ((unsigned char)1024)
#define BLACKSHIP_HEADER_LEN        ((unsigned char)3)
#define BLACKSHIP_FOOTER_LEN        ((unsigned char)1)
#define BLACKSHIP_MAX_DATA_LEN      ((unsigned char)(BLACKSHIP_MAX_MSG_LEN - (BLACKSHIP_HEADER_LEN + BLACKSHIP_FOOTER_LEN)))




int blackship_open(char* serial_port)
{
  struct termios term;
  int flags;
  int fd;

  printf("#Opening Blackship I/F [%s]\n", serial_port);

  /* Open the serial port, with O_NONBLOCK so that the open() doesn't
   * block. */
  if((fd = open(serial_port, O_RDWR | O_SYNC | O_NONBLOCK,
		S_IRUSR | S_IWUSR )) < 0) {
    printf( "open(2) failed on %s: %s", serial_port, strerror(errno));
    return(-1);
  }
  
  /* Go to blocking mode. */
  if((flags = fcntl(fd, F_GETFL)) < 0) {
    printf( "fcntl(2) failed on %s: %s", serial_port,
	    strerror(errno));
    close(fd);
    return(-1);
  }

  if(fcntl(fd, F_SETFL, flags ^ O_NONBLOCK) < 0) {
    printf( "fcntl(2) failed on %s: %s", serial_port,
	    strerror(errno));
    close(fd);
    return(-1);
  }

  /* Flush any leftover bytes on the port. */
  if(tcflush(fd, TCIFLUSH ) < 0) {
    printf( "tcflush(3) failed on %s: %s", 
	    serial_port, strerror(errno));
    close(fd);
    return(-1);
  }

  /* Retrieve current port settings. */
  if(tcgetattr(fd, &term) < 0) {
    printf( "tcgetattr(3) failed on %s: %s", 
	    serial_port, strerror(errno));
    close(fd);
    return(-1);
  }
  
  /* Make the port raw and set relevant params. */
  cfmakeraw(&term);
  cfsetispeed(&term, BLACKSHIP_BAUDRATE);
  cfsetospeed(&term, BLACKSHIP_BAUDRATE);

  /* Write the changes back to the port. */
  if(tcsetattr(fd, TCSAFLUSH, &term) < 0 ) {
    printf( "tcsetattr(3) failed on %s: %s", 
	    serial_port, strerror(errno));
    close(fd);
    return(-1);
  }

  // get first status
  int le, re;
  blackship_get_encoder(fd, le, re);

  printf("#Opening Blackship I/F done\n");

  
  return(fd);
}

int blackship_close(int fd)
{
  printf( "#Closing connection.\n");
  if(close(fd) < 0) {
    printf( "close(2) failed : %s", strerror(errno));
    return(-1);
  }

  printf( "#Closing connection done\n");
  return(0);
}

int blackship_send_message(int fd, unsigned char *msg, int len)
{
  int i;

  char ccc[1];
  
  for(i=0;i<len;i++){
    ccc[0] = msg[i];
    if( write(fd, ccc,  1) < 0) {
      printf( "write(2) failed: %s", strerror(errno));
      return(-1);
    }
  }
  return(len);
}


int blackship_recv_message(int fd, unsigned char *msg, int num)
{
  int i;
  
  char ccc[1];

  while(1){
    read(fd, ccc, 1);
    if (ccc[0] == 0x02)
      break;
  }
  msg[0] = ccc[0];
  
  for (i=1;i<num;i++){
    read(fd, ccc, 1);
    msg[i] = ccc[0];
  }

  return(num);
}




void blackship_get_encoder(int fd, int& le, int& re)
{
  
  unsigned char smsg[BLACKSHIP_MAX_DATA_LEN];
  unsigned char rmsg[BLACKSHIP_MAX_DATA_LEN];

  smsg[0] = BLACKSHIP_STX;
  smsg[1] = BLACKSHIP_CPUID_MOTOR_LEFT;
  smsg[2] = 5;
  smsg[3] = BLACKSHIP_MODE_MOTOR_REQUEST;
  smsg[4] = BLACKSHIP_ETX;

  blackship_send_message(fd, smsg, 5);

  blackship_recv_message(fd, rmsg, 10);

  re = rmsg[5] << 8 | rmsg[6];
  le = rmsg[7] << 8 | rmsg[8];
}

unsigned short prevr = 0;
unsigned short prevl = 0;
int curr = 0;
int curl = 0;

coil::TimeValue blackship_get_encoder2(int fd, int &left, int &right)
{  

  static bool first = true;

  unsigned char smsg[BLACKSHIP_MAX_DATA_LEN];
  unsigned char rmsg[BLACKSHIP_MAX_DATA_LEN];

  smsg[0] = BLACKSHIP_STX;
  smsg[1] = BLACKSHIP_CPUID_MOTOR_LEFT;
  smsg[2] = 5;
  smsg[3] = BLACKSHIP_MODE_MOTOR_REQUEST;
  smsg[4] = BLACKSHIP_ETX;

  unsigned short cur = 0;
  //---------------------------------------

  //
  coil::TimeValue tvprev = coil::gettimeofday();

  // get left
  if(blackship_send_message(fd, smsg, 5) < 0){
    printf("bs_if:send_message error\n");
    return tvprev;
  }
  blackship_recv_message(fd, rmsg, 10);

  coil::TimeValue tvpost = coil::gettimeofday();
  //double tvdiff = tvpost - tvprev;
  //printf("error tvdiff %f\n", tvdiff);
  //-------------------------
  //error about 0.0158 sec
  //-------------------------

  //-------------------------------------
  //middle time is used as timestamp
  //-------------------------------------
  coil::TimeValue time = (double)tvprev * 0.5 + (double)tvpost * 0.5;


  cur = (rmsg[7]<<8)+rmsg[8];
  
  if (first) {
    curl = (int)cur;
    prevl = cur;
  }
  
  {
    int diff = 0;
    
    if (prevl > 65535 - 10000 && cur < 10000) {
      //overflow
      diff = (int)cur + (65536 - (int)prevl);
    }
    else if (prevl < 10000 && cur > 65535 - 10000) {
      //underflow
      diff = -((int)prevl + (65536 - (int)cur));
    }
    else {
      diff = (int)cur - (int)prevl;
    }
    prevl = cur;
    curl += diff;
  
    left = curl;
  }


  cur = (rmsg[5]<<8)+rmsg[6];
  
  if (first) {
    curr = (int)cur;
    prevr = cur;
  }

  {
    int diff = 0;
    
    if (prevr > 65535 - 10000 && cur < 10000) {
      //overflow
      diff = (int)cur + (65536 - (int)prevr);
    }
    else if (prevr < 10000 && cur > 65535 - 10000) {
      //underflow
      diff = -((int)prevr + (65536 - (int)cur));
    }
    else {
      diff = (int)cur - (int)prevr;
    }
    prevr = cur;
    curr += diff;
  
    right = curr;
  }
  
  first = false;


  return time;
}



int blackship_set_speed(int fd, int ls, int rs)
{
  unsigned char smsg[BLACKSHIP_MAX_DATA_LEN];

  smsg[0] = BLACKSHIP_STX;
  smsg[1] = BLACKSHIP_CPUID_MOTOR_LEFT;
  smsg[2] = 0x07;
  smsg[3] = 0x00;
  smsg[4] = (unsigned char)ls;
  smsg[5] = 0x00;
  smsg[6] = BLACKSHIP_ETX;

  if(blackship_send_message(fd, smsg, 7) < 0)
    return(-1);

  smsg[1] = BLACKSHIP_CPUID_MOTOR_RIGHT;
  smsg[4] = (unsigned char)rs;

  if(blackship_send_message(fd, smsg, 7) < 0)
    return(-1);
  
  return(1);
}


int blackship_set_timeout(int fd, int tm)
{
  unsigned char smsg[BLACKSHIP_MAX_DATA_LEN];

  smsg[0] = BLACKSHIP_STX;
  smsg[1] = BLACKSHIP_CPUID_MOTOR_LEFT;
  smsg[2] = 0x07;
  smsg[3] = 0x07;
  smsg[4] = (unsigned char)tm;
  smsg[5] = 0x00;
  smsg[6] = BLACKSHIP_ETX;

  if(blackship_send_message(fd, smsg, 7) < 0)
    return(-1);

  smsg[1] = BLACKSHIP_CPUID_MOTOR_RIGHT;

  if(blackship_send_message(fd, smsg, 7) < 0)
    return(-1);
  
  return(1);
}

